/*
 *    DecisionTree.java
 *    Copyright (C) 1999 Pierre Geurts
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import java.util.*;

public class DecisionTree {
  /* problem definition */
  Database db;
  ObjectSet LearningSet;
  Vector CandAttributes;
  SymbAtt goal;
  
  /* parameter */
  double EntropyThres=0;
  double ScoreThres=0;
  
  /* root node */
  DtNode root;
  
  /* stack of open node */
  Stack openNodes=new Stack();

  /* Ordered attributes list for the current node */
  Vector CurrentNodeAtt;

  /**
   *  create a trivial tree with the default parameters values
   */
  DecisionTree(Database d, ObjectSet ls, Vector CandAtt, SymbAtt g) {
    db=d;
    LearningSet=ls;
    CandAttributes=CandAtt;
    goal=g;
    DtNode.nbnodes=0;
    root=new DtNode(LearningSet, g);
    root.graphnode=new WNode("open", null, null, null, null, 0, 0, 0, root);
    root.graphnode.selected=true;
    //    openNodes.push(root);
  }
  
  /**
   *  create a trivial tree with the user's parameter values
   */
  DecisionTree(Database d, ObjectSet ls, Vector CandAtt, SymbAtt g, 
	       double et, double st) {
    EntropyThres=et;
    ScoreThres=st;
    db=d;
    LearningSet=ls;
    CandAttributes=CandAtt;
    goal=g;
    DtNode.nbnodes=0;
    root=new DtNode(LearningSet, g);
    root.graphnode=new WNode("open", null, null, null, null, 0, 0, 0, root);
    root.graphnode.selected=true;
    //    openNodes.push(root);
  }
  
  /**
   *  build completely the decision tree
   */
  public void buildDt() {
    // we search this tree for all the remaining open nodes
    getOpenNodes();
    // then we build the tree
    while(!openNodes.empty()) {
      expandTopNode();
    }
  }
  
  /**
   *  expand the node on the top of stack always with the best choice
   */
  public void expandTopNode() {
    DtNode topnode=(DtNode)openNodes.pop();
    scanNode(topnode);
    Vector result=topnode.split(CurrentNodeAtt,0,goal);
    //-- we put the successors on the top of the stack (in reverse order)
    if (result!=null) {
      for (int i=result.size()-1; i>=0; i--) {
	openNodes.push(result.elementAt(i));
      }
    }
  }

  /**
   *  scan the attributes for the top node
   */
  public Vector scanNode(DtNode topnode) {
    CurrentNodeAtt=topnode.scanAttributes(CandAttributes, goal, EntropyThres, ScoreThres);
    /* build a vector of textual description of candidate tests */
    Vector result=new Vector();
    for (int i=0; i<CurrentNodeAtt.size()-1; i++)
      result.addElement(((Attribute)CurrentNodeAtt.elementAt(i)).describeBestTest());
    result.addElement(CurrentNodeAtt.lastElement());
    return result;
  }
  
  /** 
   *  split the top node according to the choice in the CurrentNodeAtt 
   *  Vector
   */
  public DtNode splitNode(DtNode topnode, int choice) {
    Vector result=topnode.split(CurrentNodeAtt,choice,goal);
    if (result!=null) {
      // we return the first successor
      return (DtNode)result.elementAt(0);
    } else {
      // otherwise, we return the leftmost open node
      return getLeftMostOpenNode();
    }
  }

  /**
   *  get all the open nodes of the tree (needed because the user can
   *  now open the node by himself)
   */
  public void getOpenNodes() {
    openNodes.removeAllElements();
    root.getOpenNodes(openNodes);
  }
  
  /**
   *  look for the leftmost open node in the tree
   *  (this is used after splitting to find the next node to split)
   */
  
  public DtNode getLeftMostOpenNode() {
    return root.getLeftMostOpenNode();
  }

  /**
   *  get a description of the current open node 
   */
  public String getNodeDescription(DtNode node) {
    return node.getNodeDescription(goal);
  }
  
  /**
   *  propagate this object across the tree and get the associated class
   */
  public String Propagate(Obj o) {
    return root.getClass(o);
  }
  
  /**
   *  test the dt on an objectset
   */
  public double testDt(ObjectSet ts) {
    String treeclass, trueclass;
    int nberrors=0;
    for (int i=0; i<ts.size; i++) {
      treeclass=Propagate(ts.elementAt(i));
      trueclass=(String)goal.Value(ts.elementAt(i));
      if (!trueclass.equals(treeclass))
	nberrors++;
    }
    System.out.println("nb of errors : " + nberrors + " out of " + ts.size);
    return 1.0*nberrors/ts.size;
  }

  /**
   *  test the tree using cross-validation 
   *     argument - the number of subsets into which
   *     the data base must be divided
   */
  public double testDtCrossvalidation(int tssize) {
    DecisionTree dt;
    int n=LearningSet.size/tssize;
    Vector tsvector=new Vector();
    ObjectSet remainingobjects=LearningSet;
    System.out.println("");
    System.out.println("Cross validation : " + n + ", LS : " 
		       + (LearningSet.size-tssize) + " , TS: " +
		       tssize);
    System.out.println("");
    
    // we construct the set of test sets
    for (int i=0; i<n; i++) {
      ObjectSet os=remainingobjects.selectRandom(tssize);
      tsvector.addElement(os);
      remainingobjects=remainingobjects.selectNotIn(os);
    }
    
    // we construct the trees
    double result=0;
    for (int i=0; i<n; i++) {
      System.out.println("");
      System.out.println("Tree number " + (i+1));
      System.out.println("");
      ObjectSet ts=(ObjectSet)tsvector.elementAt(i);
      dt=new DecisionTree(db, LearningSet.selectNotIn(ts),
			  CandAttributes, goal,
			  EntropyThres, ScoreThres);
      dt.buildDt();
      result+=dt.testDt(ts);
    }
    
    // we return the average
    return result/n;
  }

  public void printDt() {
    System.out.println(root.printNode(0));
  }
}
