/*
 *    DtNode.java
 *    Copyright (C) 1999 Pierre Geurts
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import java.util.*;

public class DtNode {
  static int nbnodes;
  ObjectSet subset;
  String type="open";        // leaf, test or open
  String majclass="default"; // the majority class in this node
  String name;
  WNode graphnode;           // for the tree visualisation

  Test test;
  Vector successors;
  
  DtNode(ObjectSet s, SymbAtt goal) {
    subset=s;
    name="node" + nbnodes++;
    ComputeInfoNode(goal);
  }

  /**
   *  Compute some information about this node (entropy, subset summary 
   *  and class maj). This should be call when the node is created.
   */
  public void ComputeInfoNode(SymbAtt goal) {
    subset.computeEntropy(goal);
    majclass=subset.getMajClass(goal);
  }
  
  /**
   *  Scan all the cancidate attributes, compute the information gain of the
   *  best score. Return a vector of attributes ordered by decreasing values 
   *  of information.
   */
  public Vector scanAttributes(Vector ca, SymbAtt goal, double entthres, double scorethres) {
    Vector result=new Vector();

    System.out.println("expanding " + getNodeDescription(goal));
    
    //-- no object in this node (this should never happen)
    if (subset.size==0) {
      result.addElement("No object in this node -> a leaf");
      return result;
    }

    //-- entropy is low enough, we can stop splitting this node
    if (subset.entropy<=entthres) {
      result.addElement("Entropy low enough (threshold=" +
			entthres + ") -> a leaf");
      return result;
    }
    
    //-- in the other case, we scan the attributes
    //    System.out.println("number of candidate attributes : " + ca.size());
    System.out.print("[");
    double score;
    int i=0,j=0;
    Attribute att;
    for (; i<ca.size(); i++) {
      System.out.print("."); // on point per attribute
      att=(Attribute)ca.elementAt(i);
      score=att.bestscore(subset,goal);
      //      System.out.println(att.name + " " + score);
      //-- the score must be greater than the threshold (forward pruning)
      if ((score*subset.size)>scorethres) {
	j=0;
	while (j<result.size() && 
	       score<=((Attribute)result.elementAt(j)).lastbestscore)
	  j++;
	if (j==result.size())
	  result.addElement(att);
	else
	  result.insertElementAt(att,j);
      }	
    }
    System.out.println("]");

    //-- we always let the possibility to user to make this node a leaf
    result.addElement("make this node a leaf");
    return result;
  }

  public Vector split(Vector scanresult, int choice, SymbAtt goal) {
    
    //-- make this node a leaf
    if (choice==scanresult.size()-1) {
      System.out.println((String)scanresult.elementAt(choice));
      type="leaf";
      buildGraphLeafNode(goal);
      return null;
    }

    //-- we take the attribute and construct the partition
    Attribute selectedatt=(Attribute)scanresult.elementAt(choice);
    Partition partition=selectedatt.bestpartition(subset,goal);
    test=partition.test;
    successors=new Vector();
    type="test";
    System.out.println("-> a test node (score : " + selectedatt.lastbestscore + 
		       " , test : " + test.TestDescription + ")");
    
    //-- we create the successor nodes
    for (int i=0; i<partition.sets.size(); i++) {
      successors.addElement(new DtNode((ObjectSet)partition.sets.elementAt(i),
				       goal));
    }
    
    //-- we create the graphnodes
    buildGraphTestNodes();
    
    //- we return the successors (open nodes)
    return successors;
  }
  
  /**
   *  expand this node and return a vector of open nodes resulting from
   *  the expansion
   */
  public Vector expand(Vector ca, SymbAtt goal, double entthres, double scorethres) {
    
    //-- if the subset is empty, this is a leaf and the class is default
    if (subset.size==0) {
      System.out.println("-> a leaf (no objects)");
      majclass="default";
      type="leaf";
      buildGraphLeafNode(goal);
      return null;
    }

    System.out.println("expanding " + getNodeDescription(goal));
    
    //-- if the entropy is lower than the threshold, then we stop the splitting
    if (subset.entropy<=entthres) { // a new leaf
      System.out.println("-> a leaf (entropy low, " + subset.entropy
			 + "<=" + entthres + ")");
      type="leaf";
      buildGraphLeafNode(goal);
      return null;
    }

    //-- otherwise, we screen all the attributes
    double bestscore=0,score;
    Attribute bestatt=null,att;
    System.out.print("[");
    for (int i=0; i<ca.size(); i++) {
      System.out.print(".");
      att=(Attribute)ca.elementAt(i);
      score=att.bestscore(subset,goal);
      if (score>bestscore) {
	bestscore=score;
	bestatt=att;
      }
    }
    System.out.println("]");
    
    //-- the best score must be greater than the threshold (stop-splitting rule)
    if ((bestscore*subset.size)<=scorethres) {
      System.out.println("-> a leaf (score too low, " 
			 + (bestscore*subset.size) + "<=" + 
			 scorethres + ")");
      type="leaf";
      buildGraphLeafNode(goal);
      return null;
    }

    //-- we construct the partition of the learning set
    Partition partition=bestatt.bestpartition(subset,goal);
    test=partition.test;
    successors=new Vector();
    type="test";
    System.out.println("-> a test node (score : " + bestscore + 
		       " , test : " + test.TestDescription + ")");

    //-- we create the successor nodes
    for (int i=0; i<partition.sets.size(); i++) {
      successors.addElement(new DtNode((ObjectSet)partition.sets.elementAt(i),
				       goal));
    }

    //-- we create the graphnodes
    buildGraphTestNodes();
    
    //-- we return the successors (open node)
    return successors;
  }
  
  /**
   *  complete the graphnode for a leaf
   */
  public void buildGraphLeafNode(SymbAtt goal) {
    graphnode.label=majclass;
    if (majclass.equals("default")) {
      graphnode.colour=0;
    } else {
      graphnode.colour=goal.values.indexOf(majclass)+1;
    }
  }

  /**
   *  build the graphnodes for the successors of this node
   */
  public void buildGraphTestNodes() {
    //-- first we put the label of this node (the size will be computed afterwards)
    graphnode.label=test.TestDescription;
    
    //-- then the first successor
    DtNode dtchild=(DtNode)successors.elementAt(0);
    dtchild.graphnode=new WNode("open", (String)test.valuesIssues.elementAt(0),
			      graphnode, null, null,0,0,0,dtchild);
    graphnode.child=dtchild.graphnode;
    WNode previousgchild=dtchild.graphnode;
    DtNode nextdtchild;
    
    //-- and finally the others
    for (int i=1; i<successors.size(); i++) {
      nextdtchild=(DtNode)successors.elementAt(i);
      nextdtchild.graphnode=new WNode("open", (String)test.valuesIssues.elementAt(i),
				      graphnode,null,null,0,0,0,nextdtchild);
      previousgchild.sibling=nextdtchild.graphnode;
      previousgchild=nextdtchild.graphnode;
    }
  }    

  /**
   *  open a node (destroy all his successors and tests)
   */
  public void open() {
    test=null;
    type="open";
    graphnode.child=null;
    graphnode.colour=0;
    graphnode.label="open";
  }

  
  /**
   *  get all the open nodes of this subtree (in a depth first way but
   *  in the oposite direction (right to left and bottom to top)
   */
  public void getOpenNodes(Stack opennodes) {
    if (type.equals("open"))
      opennodes.push(this);
    else if (successors!=null) {
      for (int i=successors.size()-1; i>=0; i--) {
	((DtNode)successors.elementAt(i)).getOpenNodes(opennodes);
      }
    }
  }
  
  /**
   *  get the leftmost open node
   */
   public DtNode getLeftMostOpenNode() {
     if (type.equals("open"))
       return this;
     else if (successors!=null) {
       for (int i=0; i<successors.size(); i++) {
	 DtNode dtn=((DtNode)successors.elementAt(i)).getLeftMostOpenNode();
	 if (dtn!=null)
	   return dtn;
       }
       return null;
     } else 
       return null;
   }

  /**
   *  return the class associated to this object by the subtree corresponding
   *  to this node
   */
  public String getClass(Obj o) {
    if (type.equals("leaf") || type.equals("open"))
      return majclass;
    else 
      return ((DtNode)successors.elementAt(test.TestIndex(o))).getClass(o);
  }

  /** 
   *  give a textual description of this node
   */
   public String getNodeDescription(SymbAtt goal) {
    String s="";
    s+=name;
    s+=" - LS : " + subset.getSummaryString(goal);
    s+=" - ENTROPY : " + subset.entropy;
    return s;
   }
  
  /**
   *  print n spaces (to pretty print the decision tree)
   */
  private String printNtab(int n) {
    String result="";
    for (int i=0; i<n;i++) {
      result+="  ";
    }
    return result;
  }
  
  /**
   *  print a textual description of the subtree corresponding
   *  to this node. 
   */
  public String printNode(int depth) {
    String result="";
    if (type.equals("leaf") || type.equals("open")) {
      result+=printNtab(depth);
      result+="return \"" + majclass + "\";\n";
    } else if (type.equals("test")) {
      for (int i=0; i<successors.size(); i++) {
	result+=printNtab(depth);
	if (i!=0)
	  result+="} else ";
	result+="if (" + test.testDescription(i) + ") {\n";
	result+=((DtNode)successors.elementAt(i)).printNode(depth+1);
      }
      result+=printNtab(depth) + "}\n";
    } else {
      System.out.println("printNode - bad node type");
      return "";
    }
    return result;
  }   
}

